require 'json'
require 'uu/os/rest/remote_client'
require 'uu/os/uesuri'
require 'uu/os/util/message_formatter'

module UU
  module OS
    module REST

      # Object for accessing result of asynchronous processes.
      class Future
        include UU::OS::Util::MessageFormatter

        # Handshake service path.
        # @private
        @@PROCEXP_PATH = 'ues/platform/process/UESProcessExplorer'

        # Maximum wait time is 30 minutes
        # @private
        @@MAXIMUM_WAIT_TIMEOUT = 1800

        # Creates new instance of Future.
        #
        # @example
        #   # init future from string
        #   future1 = UU::OS::REST::Future.new('ues:[123]:[456]')
        #
        #   # init future from UESURI
        #   proc_uri = UU::OS::UESURI.new('ues:[123]:[456]')
        #   future1 = UU::OS::REST::Future.new(proc_uri)
        #
        #   # init future from another Future
        #   future2 = UU::OS::REST::Future.new(future1)
        #
        #   # init future with default result code
        #   future1 = UU::OS::REST::Future.new(proc_uri, 'EXPORT_RESULT')
        #
        # @param [String, UU::OS::UESURI, UU::OS::REST::Future] init_data URI of asynchronous process where to look for result or another instance of Future.
        # @param [String] resultCode Default result code. This core is used when UESFuture.get is called without result code.
        def initialize(init_data, result_code = nil)

          if init_data.kind_of?UU::OS::REST::Future
            @process_uri = init_data.get_process_uri
          else
            @process_uri = UU::OS::UESURI.new(init_data)
          end

          @default_result_code = result_code
        end

        # Returns result identified by the given +result_code+.
        #
        # @example
        #
        #   # init future object
        #   future = UU::OS::REST::Future.new('ues:[123]:[456]')
        #
        #   # Get first available result of parent process in default timeout 1 minute,
        #   # method returns nil if parent process ends with no result regardless of
        #   # subprocesses, or raises exception if parent process fails.
        #   result = future.get
        #
        #   # Same as above but timeout is 2 minutes.
        #   result = future.get(120)
        #
        #   # Get result of event with code 'EXPORT_RESULT', search also all subprocess
        #   # method returns nil if parent process or any subprocess do not set result
        #   # with code 'EXPORT_RESULT'. Raises exception if event with given code is
        #   # error event.
        #   result = future.get('EXPORT_RESULT')
        #
        #   # Same as above but timeout is 2 minutes.
        #   result = future.get('EXPORT_RESULT', 120)
        #
        # @param [String] result_code Code (regular expressions may be used) of the required result. Default is +nil+ which
        #   means we are waiting for result of parent process final state. In case of explicitly given +result_code+ result
        #   is being searched also in all nested (asynchronous) subprocesses. In case of +nil+ value result is returned as
        #   soon as parent process is finished (does not wait and search for results of nested asynchronous subprocesses).
        #   Result code can be also specified in constructor. And this default value from constructor is used only when +nil+ is
        #   passed as resultCode to this method.
        # @param [Numeric] timeout Maximum wait time in seconds (default is one minute, maximum is 30 minutes)
        #
        # @return [UU::OS::UESURI] Result in form of UESURI or +nil+ if result was not provided. Throws exception in case
        #   the result has not been provided in the given timeout.
        def get(result_code = nil, timeout = 60)
          if result_code.kind_of?Numeric
            timeout = result_code
            result_code = nil
          end
          if (timeout < 0) || (timeout > @@MAXIMUM_WAIT_TIMEOUT)
            remaining_time = @@MAXIMUM_WAIT_TIMEOUT
          else
            remaining_time = timeout
          end
          wait = 0.5 # First wait will be one second (in each loop value is multiplied by 2)
          result = nil
          svc = RemoteClient.new(UU::OS::REST, @@PROCEXP_PATH)
          
          # if result_code passed to method is null, try default result code from constructor
          result_code = @default_result_code if result_code.nil?
          
          if !result_code.nil?
            data = "{\"eventCodes\":[\"#{result_code}\"]}"
          else
            data = '{}'
          end
          while true
            result = svc.post(:getProcessResult, @process_uri, data)
            if (result == 'null') || (result == '')
              # Nil result is wrongly returned as "null" string or empty string
              # (based on process entity existence)
              result = nil
            end
            if (result)
              # Result object was provided (event matching given code or event
              # representing process final state if no explicit code is given)
              result = JSON.parse(result, :symbolize_names => true)
              if result_code.nil? && (result[:state] == 'FAILED' || result[:state] == 'ERROR')
                init_data = "{\"code\":\"UU.OS/E05P02.M16\",\"errorMessages\":[{\"code\":\"UU.OS/E05P02.M16\",\"localizedMessage\":\"Asynchronous process was not successfully finished:\\n\\t#{error_message(result)}\"}]}"
                raise UU::OS::REST::ServerException.new(init_data)
              elsif !result[:errorCode].nil?
                init_data = "{\"code\":\"UU.OS/E05P02.M16\",\"errorMessages\":[{\"code\":\"UU.OS/E05P02.M17\",\"localizedMessage\":\"Expected result cannot be provided:\\n\\t#{error_message(result)}\"}]}"
                raise UU::OS::REST::ServerException.new(init_data)
              elsif !result[:result].nil?
                # Get result UESURI
                result = UU::OS::UESURI.new(result[:result])
              else
                # No UESURI -> no result
                result = nil
              end
              break
            end
            status = JSON.parse(svc.get(:getProcessProgress, @process_uri), :symbolize_names => true)
            if status[:finished] >= 0
              # We have no result, but the process is finished
              break
            end
            if remaining_time <= 0
              # We have no result, process is not finished, but we are out of time
              raise 'Result has not been provided by process in the given time.'
            end
            # Wait longer in each cycle (in order not to flood remote server)
            wait *= 2
            if (remaining_time > wait)
              remaining_time -= wait
            else
              # If next wait is greater than remaining wait time, wait for the remaining time only
              wait = remaining_time
              remaining_time = 0
            end
            sleep(wait.to_i)
          end
          result
        end
        
        # Returns underlining process uri.
        #
        # @example
        #   # get proces uri
        #   proc_uri = future.get_process_uri
        #
        # @return [UU::OS::UESURI] Process uri
        def get_process_uri
          @process_uri
        end
        
        # Returns string representation of underlining process uri.
        #
        # @example
        #   # String representation
        #   future.to_s # String 'ues:[123]:[456]'
        #
        # @return [String] 
        def to_s
          @process_uri.to_s
        end
        
        # Returns json representation of underlining process uri.
        #
        # @example
        #   # Json representation
        #   future.to_json # String '"ues:[123]:[456]"'
        #
        # @return [String]
        def to_json(options = {})
          @process_uri.to_json(options)
        end
        
        private
        
        # @private
        def method_missing(method, *args, &block) #:nodoc:
          # Delegate unknown method calling to the inner UESURI object because of backward compatibility
          if @process_uri.respond_to? method
            # log warning
            logger.warning("Script tries to use method from the UESURI object on Future object. Method name: #{method}, Back trace: #{caller.join("\n")}")
            # send event to inner UESURI
            return @process_uri.send(method, *args, &block)
          else
            # if the method is not defined either on the UESURI object
            super
          end
        end
        
        # @private
        def logger #:nodoc:
          @@log ||= UU::OS::Logger.new(self)
        end

        # @private
        def error_message(error_event) #:nodoc:
          error_code = error_event[:errorCode] || 'UU.OS/E05100.MNA'
          if error_event[:message].nil?
            "#{error_code}: Unexpected error".to_json[1..-2]
          else
            "#{error_event[:message][:code] || error_code}: #{format_message(error_event[:message][:message], *error_event[:message][:parameters])}".to_json[1..-2]
          end
        end

      end

    end
  end
end
